#ifndef WIN32
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#endif
#include <msp/core/systemerror.h>
#include "console.h"
#include "handle_private.h"

using namespace std;

namespace {

#ifdef WIN32
DWORD stream_to_sys(Console::Stream stream)
{
	switch(stream)
	{
	case INPUT: return STD_INPUT_HANDLE;
	case OUTPUT: return STD_OUTPUT_HANDLE;
	case ERROR: return STD_ERROR_HANDLE;
	default: throw invalid_argument("stream_to_sys");
	}
}

#else
termios orig_attr;
#endif

} // namespace

namespace Msp {
namespace IO {

Console::Console(Stream s):
	stream(s)
{
	mode = (stream==INPUT ? M_READ : M_WRITE);

#ifdef WIN32
	*handle = GetStdHandle(stream_to_sys(stream));
#else
	*handle = stream;

	if(stream==INPUT)
		tcgetattr(*handle, &orig_attr);
#endif

	if(stream==INPUT)
		set_events(P_INPUT);
}

Console::~Console()
{
#ifndef WIN32
	if(stream==INPUT)
		tcsetattr(*handle, TCSADRAIN, &orig_attr);
#endif
}

void Console::set_block(bool b)
{
#ifdef WIN32
	// XXX Dunno how to do this in win32
	(void)b;
#else
	int flags = fcntl(*handle, F_GETFL);
	flags = (flags&~O_NONBLOCK) | (b?0:O_NONBLOCK);
	fcntl(*handle, F_SETFL, flags);
#endif
}

void Console::set_local_echo(bool e)
{
	check_access(M_READ);

#ifdef WIN32
	DWORD m;
	GetConsoleMode(*handle, &m);
	SetConsoleMode(*handle, (m&~ENABLE_ECHO_INPUT) | (e?ENABLE_ECHO_INPUT:0));
#else
	termios t;
	tcgetattr(*handle, &t);
	t.c_lflag = (t.c_lflag&~ECHO) | (e?ECHO:0);
	tcsetattr(*handle, TCSADRAIN, &t);
#endif
}

void Console::set_line_buffer(bool l)
{
	check_access(M_READ);

#ifdef WIN32
	DWORD m;
	if(!GetConsoleMode(*handle, &m))
		throw system_error("GetConsoleMode");
	if(!SetConsoleMode(*handle, (m&~ENABLE_LINE_INPUT) | (l?ENABLE_LINE_INPUT:0)))
		throw system_error("SetConsoleMode");
#else
	// XXX ICANON does more than just set line buffering, may need a bit more thought
	termios t;
	if(tcgetattr(*handle, &t)==-1)
		throw system_error("tcgetattr");
	t.c_lflag = (t.c_lflag&~ICANON) | (l?ICANON:0);
	if(tcsetattr(*handle, TCSADRAIN, &t)==-1)
		throw system_error("tcsetattr");
#endif
}

void Console::get_size(unsigned &rows, unsigned &cols)
{
	check_access(M_WRITE);

#ifdef WIN32
	// XXX Figure out how to do this
	rows = 24;
	cols = 80;
#else
	struct winsize wsz;
	if(ioctl(*handle, TIOCGWINSZ, &wsz)==-1)
		throw system_error("ioctl TIOCGWINSZ");
	rows = wsz.ws_row;
	cols = wsz.ws_col;
#endif
}

unsigned Console::do_write(const char *buf, unsigned len)
{
	check_access(M_WRITE);

	return sys_write(handle, buf, len);
}

unsigned Console::do_read(char *buf, unsigned len)
{
	check_access(M_READ);

	unsigned ret = sys_read(handle, buf, len);
	if(ret==0)
		set_eof();

	return ret;
}

Console &Console::instance(Stream s)
{
	static Console in(INPUT);
	static Console out(OUTPUT);
	static Console err(ERROR);

	switch(s)
	{
	case INPUT: return in;
	case OUTPUT: return out;
	case ERROR: return err;
	}

	throw invalid_argument("Console::instance");
}

Console &cin = Console::instance(Console::INPUT);
Console &cout = Console::instance(Console::OUTPUT);
Console &cerr = Console::instance(Console::ERROR);

} // namespace IO
} // namespace Msp
